# FindMinimaxPolynomial

Given a high-precision implementation of an univariate real function, finds the minimax
polynomial for approximating the function over a given interval.

The minimax polynomial is the polynomial that minimizes the maximum approximation error
over all polynomials with the given degree (or set of monomials, more precisely).

The package may be useful for implementing mathematical functions for a fixed precision
arithmetic, such as the IEEE-754 32-bit or 64-bit formats, for example for implementing
the sine and cosine functions in a standard mathematical library.

The module `Minimax`, in `src/Minimax.jl`, provides a high-level interface. The main
function is `Minimax.minimax_polynomial`. Access it's help entry, after installing the
package, like so:
```
import FindMinimaxPolynomial
?FindMinimaxPolynomial.Minimax.minimax_polynomial
```

## Usage example

```julia
# We'll be using BigFloat, so increase it's precision
setprecision(BigFloat, 7 * 2^7)

import FindMinimaxPolynomial, Tulip, MathOptInterface

# Shorthands
const FMP = FindMinimaxPolynomial
const MMX = FMP.Minimax
const PPTI = FMP.PolynomialPassingThroughIntervals
const NE = FMP.NumericalErrorTypes
const to_poly = FMP.ToSparsePolynomial.to_sparse_polynomial
const mmx = MMX.minimax_polynomial
const error_type_relative = NE.RelativeError()
const MOI = MathOptInterface
const itv_max_err = FMP.ApproximateInfinityNorm.interval_max_err

function make_lp()
  lp = Tulip.Optimizer{BigFloat}()

  # Drastically increase iteration limit
  MOI.set(lp, MOI.RawOptimizerAttribute("IPM_IterationsLimit"), 2000)

  # Disable presolve
  MOI.set(lp, MOI.RawOptimizerAttribute("Presolve_Level"), 0)

  lp
end

# The monomials for an odd polynomial with degree 7
monomials_4_odd = 1:2:7

# The interval on which we'll approximate the sine
itv_sin = (big"0.1", big"0.2");

# Approximate the sine from 0.1 to 0.2 with such a polynomial
res_sin = mmx(
  make_lp,
  sin,
  (itv_sin,),
  monomials_4_odd,
);

# Convert the coefficients into a polynomial for evaluation, plotting, etc.
poly_sin = to_poly(res_sin.coefficients, monomials_4_odd)

# The monomials for an even polynomial with degree 6
monomials_4_even = 0:2:6

# The interval on which we'll approximate the cosine
itv_cos = (big"1.0", big"2.0");

# Roots don't bother us! (The cosine has a root at π/2, which is on the chosen
# interval.)
res_cos = mmx(
  make_lp,
  cos,
  (itv_cos,),
  monomials_4_even,
);

poly_cos = to_poly(res_cos.coefficients, monomials_4_even)

# Let's plot our approximation errors!

using Gadfly

err_sin(x) =
  let x = BigFloat(x)
    abs(1 - poly_sin(x)/sin(x))
  end

err_sin_max_const = itv_max_err(
  NE.RelativeError(),
  sin,
  poly_sin,
  itv_sin,
).max_err

err_sin_max(x) = err_sin_max_const

plot([err_sin, err_sin_max], Float64.(itv_sin)...)

err_cos(x) =
  let x = BigFloat(x)
    abs(1 - poly_cos(x)/cos(x))
  end

err_cos_max_const = itv_max_err(
  NE.RelativeError(),
  cos,
  poly_cos,
  itv_cos,
).max_err

err_cos_max(x) = err_cos_max_const

plot([err_cos, err_cos_max], Float64.(itv_cos)...)
```

## Background

See these topics for some background (I link to Wikipedia articles, but a Web search
might be a better option, if necessary):

* [minimax approximation algorithm](https://en.wikipedia.org/wiki/Minimax_approximation_algorithm)

* [uniform norm](https://en.wikipedia.org/wiki/Uniform_norm), also known as *sup norm*,
  *supremum norm*, *Chebyshev norm*, *infinity norm*, or (in this case) *max norm* or
  *maximum norm*

* [equioscillation theorem](https://en.wikipedia.org/wiki/Equioscillation_theorem)

### Remez

Some links regarding the Remez algorithm. It's not used in this software, but it's still
the standard algorithm people reach for, even though using Remez, or Remez-based tools,
successfully is something of a "black art":

* [Wikipedia article](https://en.wikipedia.org/wiki/Remez_algorithm)

* [article on some 'blog](https://xn--2-umb.com/22/approximation/)

* [some PDF document](http://www.jaeckel.org/IndustrygradeFunctionApproximation.pdf)

* [Boost Remez docs](https://www.boost.org/doc/libs/1_81_0/libs/math/doc/html/math_toolkit/remez.html),
  [Boost docs](https://www.boost.org/doc/libs/1_81_0/libs/math/doc/html/math_toolkit/internals/minimax.html)
  for a Boost-internal tool based on Remez

Sollya is, to my knowledge, the state-of-the-art Remez-based tool:

* [homepage](https://sollya.org)

* [Git repo](https://gitlab.inria.fr/sollya/sollya), on INRIA's Gitlab

* [mailing list for users](https://sympa.inria.fr/sympa/arc/sollya.users)

* [development mailing list](https://sympa.inria.fr/sympa/arc/sollya.development)

## How is this implemented?

Doesn't use the usual Remez algorithm, which is great, because my algorithm seems to work
for a much wider class of functions than Remez. In fact, I think `minimax_polynomial`
should be able (not necessarily on the first try, because of the randomization) to find a
minimax polynomial for any function for which this is possible at all.

The core idea is to formulate a linear programming (LP) problem that finds a minimax
polynomial of fixed degree over a *finite* set of points from the target interval (the
`PolynomialPassingThroughIntervals` module does this part).  Then, since the polynomial
we're looking for is of fixed degree, adding a point to the set and repeating the LP
optimization iteratively should eventually produce a polynomial arbitrarily close to the
minimax polynomial over the *entire interval*.

A specific improvement over the Remez algorithm, when optimizing the relative error, is
that roots of the function that's being approximated don't present much of a problem to
this algorithm. A root, of course, presents a severe restriction on the minimax
polynomials that might be found, but my algorithm takes care of any possible roots on
its own, automatically. This is unlike with Remez, where roots, if supported at all for
relative error, must be given in advance exactly.

## Further work

A work-in-progress Julia package that will use this package to actually implement the
polynomial approximations is currently located here:
https://gitlab.com/nsajko/PolynomialApproximation.jl
