# Copyright © 2023 Neven Sajko

module ApproximateInfinityNorm

import
  Polynomials,
  ..NumericalErrorTypes,
  ..NumericalErrors

const NE = NumericalErrorTypes
const numerical_error = NumericalErrors.numerical_error

is_tiny(::NE.AbsoluteError, ::Real, ::Real) = false

is_tiny(::NE.RelativeError, accurate::Real, maximal_tiny::Real) =
  abs(accurate) < maximal_tiny

struct SortedBoundedLengthVector{T <: Any}
  max_len::Int

  # Sorted in descending order
  sorted::Vector{T}

  function SortedBoundedLengthVector{T}(max_len::Int) where {T <: Any}
    v = T[]
    sizehint!(v, max_len)
    new{T}(max_len, v)
  end
end

Base.isempty(v::SortedBoundedLengthVector) = isempty(v.sorted)

function linear_search(v::SortedBoundedLengthVector{T}, e::T) where {T}
  index = firstindex(v.sorted)
  while (index ≤ lastindex(v.sorted)) && !isless(v.sorted[index], e)
    index += true
  end
  index
end

Base.length(v::SortedBoundedLengthVector) = Base.length(v.sorted)

Base.last(v::SortedBoundedLengthVector) = Base.last(v.sorted)

# Pops the smallest element
Base.pop!(v::SortedBoundedLengthVector) = Base.pop!(v.sorted)

is_full(v::SortedBoundedLengthVector) = v.max_len == length(v)

function Base.push!(v::SortedBoundedLengthVector{T}, e::T) where {T}
  index = linear_search(v, e)
  after = index == lastindex(v.sorted) + true
  full = is_full(v)
  (!after & full) && pop!(v)
  (!after | !full) && insert!(v.sorted, index, e)
  v
end

Base.sort!(
  v::SortedBoundedLengthVector{T};
  lt = isless,
  by = identity,
  rev::Bool = false,
) where {T} =
  Base.sort!(v.sorted, lt = lt, by = by, rev = rev)

struct Node{F <: Real}
  # Interval bounds
  interval::NTuple{2, F}

  # Max error among the errors of the upper and lower interval bounds
  error::F
end

Base.isless(l::Node{F}, r::Node{F}) where {F} = Base.isless(l.error, r.error)

successor_density(density::S, depth_ratio::Real) where {S <: Signed} =
  let t = floor(S, density * depth_ratio)
    # It is not allowed for a node to have one child
    isone(t) ? zero(S) : t
  end

# The height of a node is the length of the longest downward path to a leaf from that
# node. The height of the root is the height of the tree.
#
# The `density` is the number of children of the root node.
#
# It is not allowed for a node to have one child.
function tree_height(density::Int, depth_ratio::Real)
  height = zero(Int)
  while false < density
    height += true
    density = successor_density(density, depth_ratio)
  end
  height
end

function tree_densities(density::Int, depth_ratio::Real)
  ret = Int32[]
  sizehint!(ret, tree_height(density, depth_ratio))
  while false < density
    push!(ret, density)
    density = successor_density(density, depth_ratio)
  end
  ret
end

struct ErrorWithStatus{F <: Real}
  is_tiny::Bool
  error::F
end

function numerical_error_with_status(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  position::F,
  maximal_tiny::Real,
) where {
  Fun <: Any,
  F <: Real,
}
  accurate = fun(position)
  approximate = evalpoly(position, pol)
  ErrorWithStatus{F}(
    is_tiny(error_type, accurate, maximal_tiny),
    numerical_error(error_type, accurate, approximate),
  )
end

function interval_max_err(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  interval::NTuple{2, F},
  maximal_tiny::Real = 0.00001;
  density::Int = 1200,
  breadth_limit::Int = 5,
  depth_ratio::Real = 1/4,
) where {
  Fun <: Any,
  F <: Real,
}
  # Some kind of tree traversal, where the structure of the tree is determined
  # implicitly by the density, limit and ratio parameters. The elements of the tree are
  # sub-intervals.

  # TODO: randomly perturb

  N = Node{F}
  V = SortedBoundedLengthVector{N}

  densities_by_level = tree_densities(density, depth_ratio)

  D = eltype(densities_by_level)

  height = length(densities_by_level)

  limited_densities_by_level =
    D[min(d, D(breadth_limit)) for d ∈ @view densities_by_level[begin:(end - 1)]]

  vectors_by_level = vcat(
    V[V(one(Int))],
    V[V(Int(d)) for d ∈ limited_densities_by_level],
    V[V(zero(Int))],
  )

  # Maximal error
  max_err = -one(F)

  # Maximal error among points where the value of the function is not "tiny". For
  # relative error, values with tiny magnitude are "tiny", while no values are "tiny"
  # when looking for absolute error.
  max_err_nontiny = -one(F)

  # Position of the "nontiny" maximal error
  position = zero(F)

  level = one(Int)
  push!(first(vectors_by_level), N(interval, -one(F)))
  while true
    if isempty(vectors_by_level[begin + level - true])
      level -= true
      iszero(level) && break
      continue
    end

    itv = pop!(vectors_by_level[begin + level - true]).interval

    boundary_errors_with_status = map(
      let f = fun, p = pol, mt = maximal_tiny
        i -> numerical_error_with_status(error_type, f, p, i, mt)
      end,
      itv,
    )

    dens_index = firstindex(densities_by_level) + level - true

    dens = Int(densities_by_level[dens_index])

    itv_len = last(itv) - first(itv)

    point_old = first(itv)
    err_old = first(boundary_errors_with_status).error

    # Process initial point
    let err_sta = first(boundary_errors_with_status),
        err = err_sta.error, val_is_tiny = err_sta.is_tiny,
        point = first(itv)
      (max_err < err) && (max_err = err)
      if (max_err_nontiny < err) & !val_is_tiny & !(point ∈ interval)
        max_err_nontiny = err
        position = point
      end
    end

    for point_index ∈ one(Int):(dens - true)
      point = first(itv) + itv_len * (point_index/dens)
      val_accur = fun(point)
      val_approx = evalpoly(point, pol)
      val_is_tiny = is_tiny(error_type, val_accur, maximal_tiny)
      err = numerical_error(error_type, val_accur, val_approx)

      (max_err < err) && (max_err = err)
      if (max_err_nontiny < err) & !val_is_tiny & !(point ∈ interval)
        max_err_nontiny = err
        position = point
      end

      node_interval = N((point_old, point), max(err_old, err))
      push!(vectors_by_level[begin + level], node_interval)

      point_old = point
      err_old = err
    end

    # Process final point and final interval/child node
    let err_sta = last(boundary_errors_with_status),
        err = err_sta.error, val_is_tiny = err_sta.is_tiny,
        point = last(itv)
      (max_err < err) && (max_err = err)
      if (max_err_nontiny < err) & !val_is_tiny & !(point ∈ interval)
        max_err_nontiny = err
        position = point
      end

      node_interval = N((point_old, point), max(err_old, err))
      push!(vectors_by_level[begin + level], node_interval)
    end

    level += true
  end

  (max_err = max_err, max_err_nontiny = max_err_nontiny, position = position)
end

end
