# Copyright © 2023 Neven Sajko

module Minimax

import
  MathOptInterface,
  Polynomials,
  ..NumericalErrorTypes,
  ..NumericalErrors,
  ..PolynomialPassingThroughIntervals,
  ..Perturbed,
  ..ApproximateInfinityNorm,
  ..IntervalHelpers,
  ..ToSparsePolynomial

const MOI = MathOptInterface
const NE = NumericalErrorTypes
const PPTI = PolynomialPassingThroughIntervals
const ItvH = IntervalHelpers
const numerical_error = NumericalErrors.numerical_error
const interval_max_err = ApproximateInfinityNorm.interval_max_err

# Should be used as a supertype for callables
const Callable = Any

check_val_tiny(::NE.AbsoluteError, ::F, ::F) where {F <: Real} = nothing

function check_val_tiny(::NE.RelativeError, val::F, maximal_tiny::F) where {F <: Real}
  (abs(val) < maximal_tiny) && error("value too tiny")
  nothing
end

function check_valid_lp_grid(
  endpoint::F,
  fun::Fun,
  error_type::NE.NumericalErrorType,
  maximal_tiny::F,
) where {F <: Real, Fun <: Callable}
  val = fun(endpoint)
  isfinite(val) || error("interval endpoint not finite")
  check_val_tiny(error_type, val, maximal_tiny)
  nothing
end

check_valid_lp_grid(
  fun::Fun,
  error_type::NE.NumericalErrorType,
  maximal_tiny::Real,
) where {Fun <: Callable} =
  let t = maximal_tiny
    ep -> check_valid_lp_grid(ep, fun, error_type, t)
  end

max_err_on_grid(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  poly::Polynomials.AbstractPolynomial{F},
  grid::Vector{F},
) where {
  Fun <: Callable,
  F <: AbstractFloat,
} =
  mapreduce(
    let p = poly
      x -> numerical_error(error_type, fun(x), evalpoly(x, p))
    end,
    max,
    grid,
    init = zero(F),
  )

struct BadSegment{F <: Real}
  outer_index::Int
  inner_index::Int

  # Position within the segment that a new grid point should be placed to if this
  # segment is chosen
  position::F
end

const NonemptyNTuple = Tuple{T, Vararg{T, n}} where {n, T}
const TupleOrVector = Union{Vector{T}, (NonemptyNTuple{n, T} where {n})} where {T}

function minimax_polynomial_impl(
  make_lp::LP,
  fun::Fun,
  ret_grids::TupleOrVector{Vector{F}},
  monomials::AbstractVector{<:Integer},
  error_type::NE.NumericalErrorType,
  ::LPOption;
  max_error_upper_bound::F,
  worst_segments_density::Int,
  worst_segments_breadth_limit::Int,
  worst_segments_depth_ratio::Real,
  exit_condition::Real,
  maximal_tiny::Real,
) where {
  LP <: Callable,
  Fun <: Callable,
  F <: AbstractFloat,
  LPOption <: PPTI.FeasibleOption,
}
  # * Note: when calculating relative error, first check whether the arguments
  #   are equal and return zero if so. Otherwise there would be problems when
  #   both arguments are zero.
  #
  # * Handling singularities:
  #   * The function could evaluate to a value that is not finite, which causes
  #     an error to be thrown
  #   * If the function evaluates to zero or to a small value, the error could
  #     evaluate to infinity or to a very large number, which we check by
  #     comparing the function value with the maximal "tiny" value
  #     * In the case of absolute error, this will not happen
  #     * In the case of relative error:
  #       * If the point is one of the two edges of the interval, this is an
  #         error
  #       * If the point is meant for computing relative error, just compute
  #         the error, as long as it's not NaN it's a valid error. However, now
  #         it's very likely that we'll be putting an LP point somewhere around
  #         here...
  #       * If the point is meant for LP, we need to find another point close
  #         by, because tiny intervals could cause numerical issues for LP.

  (false ≤ max_error_upper_bound) ||
    error("error can not be negative: $max_error_upper_bound")
  (20 < worst_segments_density) ||
    error("density too small: worst_segments_density")
  (false < worst_segments_breadth_limit) ||
    error("breadth limit must be positive")
  (false < worst_segments_depth_ratio < true) ||
    error("ratio must be between zero and one: $worst_segments_depth_ratio")
  (false ≤ exit_condition) ||
    error("error can not be negative: $exit_condition")
  (false ≤ maximal_tiny) ||
    error("magnitude can not be negative: $maximal_tiny")

  isempty(ret_grids) && error("no grids given")

  (mapreduce(length, +, ret_grids, init = 0) < length(monomials) + 1) &&
    error("grid too small")

  # Check that all points on each grid are finite and, if we're optimizing relative
  # error, that the value isn't too tiny
  foreach(
    let mt = maximal_tiny
      g -> foreach(
        check_valid_lp_grid(fun, error_type, maximal_tiny),
        g,
      )
    end,
    ret_grids,
  )

  # Check that:
  #
  #   * each grid is sorted
  #   * each grid contains at least two elements
  #   * each grid contains no duplicates 
  foreach(
    function(g::Vector{F})
      issorted(g) || error("grid not sorted")
      (2 ≤ length(g)) || error("grid has too little elements")
      allunique(g) || error("grid has duplicates")
      nothing
    end,
    ret_grids,
  )

  # Check that the collection of grids is sorted and the grids are discontiguous.
  for i ∈ eachindex(ret_grids)[begin:(end - 1)]
    (last(ret_grids[i]) < first(ret_grids[i + 1])) ||
      error("grids overlapping or not in order")
  end

  ret_coefs = F[]
  ret_lp = make_lp()

  # Something like this:
  #
  # 1. Do LP with the grid. Exit if there's no solution.
  # 2. Test the resulting polynomial's error with a denser grid
  # 3. Pick a segment between some LP grid points such that the maximal error
  #    on the segment is greater than the max error on LP grid
  # 4. Add a point for LP within the chosen segment
  # 5. Go back to step 1

  while true
    lp_grid = reduce(vcat, ret_grids, init = F[])

    func_vals = map(fun, lp_grid)

    lp_intervals = ItvH.intervals_given_max_err(
      func_vals,
      error_type,
      max_error_upper_bound,
    )

    # LP
    ret_coefs = PPTI.polynomial_passing_through_intervals!(
      ret_lp,
      monomials,
      lp_grid,
      lp_intervals,
      F,
      PPTI.MinimaxOption{LPOption}(func_vals, error_type, max_error_upper_bound),
    )

    # Exit if the LP failed
    isempty(ret_coefs) && break

    polynomial = ToSparsePolynomial.to_sparse_polynomial(ret_coefs, monomials)

    max_lp_err = max_err_on_grid(error_type, fun, polynomial, lp_grid)

    max_max_err = -one(F)

    worst_segments = (
      finite_error = BadSegment{F}[],
      infinite_error = BadSegment{F}[],
    )

    # At the same time:
    #   * Look for max max error
    #   * Look for bad segments
    for segment_index_outer ∈ eachindex(ret_grids),
        segment_index ∈ eachindex(ret_grids[segment_index_outer])[begin:(end - 1)]
      segment = map(
        let g = ret_grids[segment_index_outer]
          i -> g[i]
        end,
        (segment_index, segment_index + 1),
      )
      errs_with_position = interval_max_err(
        error_type,
        fun,
        polynomial,
        segment,
        maximal_tiny,
        density = worst_segments_density,
        breadth_limit = worst_segments_breadth_limit,
        depth_ratio = worst_segments_depth_ratio,
      )
      err = errs_with_position.max_err
      ent = errs_with_position.max_err_nontiny
      pos = errs_with_position.position

      # Update max_max_err as necessary
      (max_max_err < err) && (max_max_err = err)

      # Update worst_segments as necessary
      if max_lp_err ≤ err
        segm = BadSegment{F}(segment_index_outer, segment_index, pos)
        if err == ent
          push!(worst_segments.finite_error, segm)
        else
          push!(worst_segments.infinite_error, segm)
        end
      end
    end

    all(isfinite, (max_max_err, max_lp_err)) && (
      (max_max_err ≤ max_lp_err) ||
      (
        numerical_error(
          NE.RelativeError(),
          max_max_err,
          max_lp_err,
        ) < exit_condition
      )
    ) &&
      # Success!
      break

    chosen_bad = BadSegment{F}[]
    if isempty(first(worst_segments))
      chosen_bad = last(worst_segments)
    elseif isempty(last(worst_segments))
      chosen_bad = first(worst_segments)
    else
      chosen_bad = worst_segments[begin + rand(Bool)]
    end

    chosen_segment = chosen_bad[rand(eachindex(chosen_bad))]

    ret_grid = ret_grids[chosen_segment.outer_index]

    # Take a point from the chosen segment
    #   1. Take a perturbed equidistant grid on the segment
    #   2. Evaluate the error on each point of the grid
    #   3. Among the non-tiny points, take the one with maximal error. If all
    #      points are tiny, error.

    # Add the chosen point to the LP grid
    insert!(ret_grid, chosen_segment.inner_index + 1, chosen_segment.position)

    ret_lp = make_lp()
  end

  (
    coefficients = ret_coefs,
    grids = ret_grids,
    lp = ret_lp,
  )
end

"""
Finds a minimax polynomial.

Arguments:

  * A function that returns a new MOI.AbstractOptimizer

    * A new MathOptInterface LP optimizer each time it's called
    * Only one will be used at a time
    * Tested with Tulip

  * A function to be approximated as input

    * Only univariate real functions are supported

  * A tuple or vector of `NTuple{2, F}` elements, each of which represents an interval
    over which the given function is to be approximated. The error is minimized over all
    intervals simultaneously.

    * `F` subtypes AbstractFloat
    * The first value in an interval is the lower bound of the interval, and the second
      value is the upper bound
    * None of the two bounds of an interval may be a root of the given
      function when the error type is relative error. If this is a problem,
      the user should try slightly shrinking or expanding the interval over
      the root.

  * The indices denoting the monomials that the sought-after polynomial is
    supposed to have

    * For example, use `0:3` to look for any polynomial of degree three, or
      `0:2:8` for any even polynomial of degree eight
    * Note: the user should use even polynomials for approximating even
      functions and odd polynomials for approximating odd functions

  * The type of the error that we're minimizing, absolute or relative

    * Optional argument

  * Options for the LP formulation

    * Optional argument

  * The upper bound on allowed maximum error

    * Keyword argument

  * Exit condition

    * Keyword argument

  * Maximal value considered to be too "tiny" for LP

    * Keyword argument

Returns a named tuple with these fields:

  * A vector, `coefficients`:

    * If nonempty, it's the coefficients of the polynomial in the standard
      basis

    * If empty, the last LP solve failed. This might mean that more monomials are needed
      or that the arithmetic precision is too low and should be increased to enable the
      algorithm to complete successfully.

  * The final LP grid, `grid`:

    * May be used for refining the result further, e.g., by running again
      with an increased segment grid point count value
    * May be used for finishing where the algorithm left off after a failure
      caused by a lack of precision (increase the precision first, though)

  * The most recently used LP optimizer, `lp`

See the package help entry for a usage example:
```
import FindMinimaxPolynomial
?FindMinimaxPolynomial
```
"""
function minimax_polynomial(
  make_lp::LP,
  fun::Fun,
  intervals::TupleOrVector{NTuple{2, F}},
  monomials::AbstractVector{<:Integer},
  error_type::NE.NumericalErrorType = NE.RelativeError(),
  ::LPOption = PPTI.ReplacedVariablesOption{:determinants}();
  max_error_upper_bound::F = typemax(F),
  initial_perturb_factor::Real = 1//16,
  worst_segments_density::Int = 2^10,
  worst_segments_breadth_limit::Int = 4,
  worst_segments_depth_ratio::Real = 1/4,
  exit_condition::Real = 2^-500,
  maximal_tiny::Real = floatmin(F)^(1//64),
) where {
  LP <: Callable,
  Fun <: Callable,
  F <: AbstractFloat,
  LPOption <: PPTI.FeasibleOption,
}
  isempty(intervals) && error("no intervals given")
  minimax_polynomial_impl(
    make_lp,
    fun,
    map(
      let pf = initial_perturb_factor, l = length(monomials) + 1
        i -> Perturbed.perturbed_grid_on_interval(l, i, pf)
      end,
      intervals,
    ),
    monomials,
    error_type,
    LPOption(),
    max_error_upper_bound = max_error_upper_bound,
    worst_segments_density = worst_segments_density,
    worst_segments_breadth_limit = worst_segments_breadth_limit,
    worst_segments_depth_ratio = worst_segments_depth_ratio,
    exit_condition = exit_condition,
    maximal_tiny = maximal_tiny,
  )
end

end
